/*
 *  linux/fs/cse451/inode.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 *
 *  Copyright (C) 1996  Gertjan van Wingerde    (gertjan@cs.vu.nl)
 *	Minix V2 fs support.
 *
 *  Modified for 680x0 by Andreas Schwab
 */

#include <linux/module.h>

#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/stat.h>
#include <linux/locks.h>
#include <linux/init.h>
#include <linux/smp_lock.h>
#include <linux/highuid.h>

#include <asm/system.h>
#include <asm/bitops.h>

#include "../cse451fs.h"

/*
 * cse451 directories can handle most operations...
 */
struct inode_operations cse451_dir_inode_operations = {
	create:		cse451_create,
	lookup:		cse451_lookup,
	link:		cse451_link,
	unlink:		cse451_unlink,
	//symlink:	cse451_symlink,
	mkdir:		cse451_mkdir,
	rmdir:		cse451_rmdir,
	mknod:		cse451_mknod,
	rename:		cse451_rename,
};

///////////////////////////////////////////////////////
// Helper functions                                  //
///////////////////////////////////////////////////////

struct inode * cse451_new_inode(const struct inode * dir, int * error)
{
        struct super_block*     sb;
	struct cse451_sb_info*	pSBInfo;
	struct cse451_inode_info* pInodeInfo;
        struct inode*           inode;
        struct buffer_head*     bh;
        int			inum;

        sb = dir->i_sb;
        inode = new_inode(sb);
        if (!inode) {
                *error = -ENOMEM;
                return NULL;
        }

        bh = NULL;
        *error = -ENOSPC;

        lock_super(sb);

	pSBInfo = (struct cse451_sb_info*)sb->u.generic_sbp;

        for (inum = 0; inum < pSBInfo->pSB->s_nNumInodes; inum++) {
                if (!test_bit(inum,pSBInfo->pImap)) break;
        }

        if (inum >= pSBInfo->pSB->s_nNumInodes) {
                iput(inode);
                unlock_super(sb);
                return NULL;
        }
        if (__test_and_set_bit(inum,pSBInfo->pImap)) { /* shouldn't happen */
                printk("new_inode (%d): bit already set", inum+1);
                iput(inode);
                unlock_super(sb);
                return NULL;
        }

	pSBInfo->pSB->s_nBusyInodes++;
        sb->s_dirt = 1;

        inode->i_uid = current->fsuid;
        inode->i_gid = (dir->i_mode & S_ISGID) ? dir->i_gid : current->fsgid;
        inode->i_ino = inum+1;	// inodes have to be named starting at 1
        inode->i_mtime = inode->i_atime = inode->i_ctime = CURRENT_TIME;
        inode->i_blocks = 0;
	inode->i_blksize = sb->s_blocksize;

	DEBUGPRINTK (("_new_inode: ino = %d\n", inum+1))

	inode->u.generic_ip = kmalloc(sizeof(struct cse451_inode_info), GFP_KERNEL);
	pInodeInfo = (struct cse451_inode_info*) inode->u.generic_ip;
	for (inum=0; inum<CSE451_NUMDATAPTRS; inum++) {
		pInodeInfo->i_datablocks[inum] = 0;	// mark not alloc'ed
	}

        insert_inode_hash(inode);
        mark_inode_dirty(inode);

        unlock_super(sb);
        *error = 0;
        return inode;
}

///////////////////////////////////////////////////////
// Callbacks from VFS                                //
///////////////////////////////////////////////////////

// 
// get rid of this guy permanently
//
void cse451_delete_inode(struct inode *inode)
{
	struct cse451_sb_info*	pSBInfo;
	unsigned long 		ino;

	DEBUGPRINTK (("_delete_inode\n"))
	DEBUGPRINTK (("... 0x%p %ld\n", inode, inode?inode->i_ino:0))

	ino = inode->i_ino;

	pSBInfo = (struct cse451_sb_info*)inode->i_sb->u.generic_sbp;

	if (ino < 1 || ino > pSBInfo->pSB->s_nNumInodes) {
		printk("cse451_delete_inode: inode 0 or nonexistent inode\n");
		return;
	}

	pSBInfo = (struct cse451_sb_info*)inode->i_sb->u.generic_sbp;

	lock_kernel();
	inode->i_size = 0;
	cse451_truncate(inode);

	if (!test_and_clear_bit(ino-1, pSBInfo->pImap)) {
		printk("free_inode: bit %lu already cleared!?\n",ino);
	} else {
		pSBInfo->pSB->s_nBusyInodes--;
		inode->i_sb->s_dirt = 1;
	}

	mark_buffer_dirty(pSBInfo->sb_bh);

	clear_inode(inode);

	unlock_kernel();

}

// 
// We're goint to remove the VFS inode from memory
//   Last chance to "clean up" anything that needs cleaning up
//

void cse451_clear_inode(struct inode *inode)
{
	struct cse451_inode_info* pInodeInfo;
	struct cse451_sb_info*	  pSBInfo;
	int ino;

	ino = inode->i_ino;

	DEBUGPRINTK (("_clear_inode\n"))
	DEBUGPRINTK (("...%d i_sb: 0x%p\n", ino, inode->i_sb))

	pSBInfo = (struct cse451_sb_info*)inode->i_sb->u.generic_sbp;

	// put_super has already wiped out pSBInfo for the root inode,
	// so careful...

	if (!ino || (pSBInfo && ino > pSBInfo->pSB->s_nNumInodes)) {
		printk("clear_inode: inode %d out of range\n",ino);
		return;
	}

	pInodeInfo = (struct cse451_inode_info*)inode->u.generic_ip;
	if (pInodeInfo) {
		kfree (pInodeInfo);
		inode->u.generic_ip = NULL;
	}

	DEBUGPRINTK (("_clear_inode returning...\n")) 
}

//
// make sure inode is in memory, and set up to go
//

void cse451_read_inode(struct inode * inode)
{
        int block, ino;
        struct buffer_head*         bh;
        struct cse451_inode*        raw_inode;
	struct cse451_super_block*  pSB;
	struct cse451_inode_info*   pInodeInfo;

        ino = inode->i_ino;

	pSB = ((struct cse451_sb_info*)(inode->i_sb->u.generic_sbp))->pSB;

        if (ino > pSB->s_nNumInodes) {
                printk("read_inode: Bad inode number on dev %s: %d is out of range\n",
                        kdevname(inode->i_dev), ino);
                return;
        }

        block = pSB->s_nInodeStart + (ino-1)/CSE451_INODES_PER_BLOCK;

	DEBUGPRINTK (("_read_inode\n"))
	DEBUGPRINTK (("...%d block=%d\n", ino, block))

        if (!(bh=bread(inode->i_dev, block, BLOCK_SIZE))) {
                printk("Major problem: unable to read inode %d from dev "
                       "%s\n", ino, kdevname(inode->i_dev));
                return;
        }
        raw_inode = ((struct cse451_inode *) bh->b_data) +
                                   (ino-1)%CSE451_INODES_PER_BLOCK;
        inode->i_mode = raw_inode->i_mode;
        inode->i_nlink = raw_inode->i_nlinks;
        inode->i_size = raw_inode->i_filesize;
	inode->i_blksize = inode->i_sb->s_blocksize;
        inode->i_blocks = 0;
        inode->i_mtime = inode->i_atime = inode->i_ctime = 0;  // not implemented...

	pInodeInfo = kmalloc(sizeof(struct cse451_inode_info), GFP_KERNEL);
	if (!pInodeInfo) {
		printk ("Kernel out of memory\n");
		return;
	}
	inode->u.generic_ip = pInodeInfo;

	DUMPX4 (inode->i_mode);
	DUMPX4 (inode->i_nlink);
	DUMPX4 (inode->i_size);
	DUMPX4 (inode->i_blocks);

        for (block = 0; block < CSE451_NUMDATAPTRS; block++)
              pInodeInfo->i_datablocks[block] = raw_inode->i_datablocks[block];

        if (S_ISREG(inode->i_mode)) {
                inode->i_op = &cse451_file_inode_operations;
                inode->i_fop = &cse451_file_operations;
                inode->i_mapping->a_ops = &cse451_aops;

        } else if (S_ISDIR(inode->i_mode)) {
                inode->i_op = &cse451_dir_inode_operations;
                inode->i_fop = &cse451_dir_operations;

        } else if (S_ISLNK(inode->i_mode)) {
		// this can't happen in cse451 base configuration...
                // inode->i_op = &page_symlink_inode_operations;
                inode->i_mapping->a_ops = &cse451_aops;

        } else
               init_special_inode(inode, inode->i_mode, raw_inode->i_datablocks[0]);

        brelse(bh);
}

// 
// update inode image on disk
//

void cse451_write_inode(struct inode * inode, int wait)
{
        struct buffer_head*	bh;
	struct cse451_inode*	raw_inode;
	struct cse451_sb_info*	pSBInfo;
	struct cse451_inode_info* pInodeInfo;
	int			ino, block;

        lock_kernel();

	pSBInfo = (struct cse451_sb_info*)inode->i_sb->u.generic_sbp;

	ino = inode->i_ino;
	if (!ino || ino > pSBInfo->pSB->s_nNumInodes) {
		printk("write_inode: Bad inode number on dev %s"
		       ": %d is out of range\n",
			kdevname(inode->i_dev), ino);
		unlock_kernel();
		return;
	}

	DEBUGPRINTK (("_write_inode: %d\n", ino))

	// have to get the full block that this inode lives in...

	block = pSBInfo->pSB->s_nInodeStart + (ino-1)/CSE451_INODES_PER_BLOCK;
	if (!(bh=bread(inode->i_dev, block, BLOCK_SIZE))) {
		printk("cse451fs: unable to read i-node block(ino=%d)\n",ino);
		unlock_kernel();
		return;
	}

	DEBUGPRINTK (("... block 0x%08x\n", block))

	raw_inode = ((struct cse451_inode *)bh->b_data) +
			(ino-1) % CSE451_INODES_PER_BLOCK;

	raw_inode->i_mode = inode->i_mode;
	raw_inode->i_nlinks = inode->i_nlink;
	raw_inode->i_filesize = inode->i_size;

	if (S_ISCHR(inode->i_mode) || S_ISBLK(inode->i_mode))
		raw_inode->i_datablocks[0] = kdev_t_to_nr(inode->i_rdev);
	else {
		DEBUGPRINTK (("... write all data ptrs\n"))
		pInodeInfo = (struct cse451_inode_info*)inode->u.generic_ip;
		for (block = 0; block < CSE451_NUMDATAPTRS; block++)
			raw_inode->i_datablocks[block] =
				pInodeInfo->i_datablocks[block];
	}

        unlock_kernel();

	mark_buffer_dirty(bh);
        brelse(bh);
}
